"""
    BodegaConnect is part of the hellofresh_data module
    meant for returning a bodega connection object for
    DB quering purposes. Note, if no variables are passed
    into the object a default, read only, user connection is returned.
"""
import os
import psycopg2
import yaml
from hellofresh_data import parameter_store

HOST_PATH = '/Prod/BodegaDB/host'
USER_PATH = '/Prod/BodegaDB/ro_user'
PASSWORD_PATH = '/Prod/BodegaDB/ro_password'

class BodegaConnect(object):
    """
        Base class for BodegaConnection
    """
    def __init__(self, bodega_user=None, bodega_password=None):

        self._conn = None
        self._bodega_user = None
        self._bodega_password = None

        self._bodega_host = \
        parameter_store.get_parameter_store_value(HOST_PATH)

        self.set_connection_credentials(bodega_user, bodega_password)


    def set_connection_credentials(self, bodega_user, bodega_password):
        """
            If no variables are passed into the object a default, read only,
            user connection is returned.
        """
        if bodega_user is None and bodega_password is  None:
            self._bodega_user = \
            parameter_store.get_parameter_store_value(USER_PATH)

            self._bodega_password = \
            parameter_store.get_parameter_store_value(PASSWORD_PATH)
        else:
            self._bodega_user = bodega_user
            self._bodega_password = bodega_password

    def bodega_connect(self):
        """
            Connecting to Bodega
        """
        try:
            self._conn = psycopg2.connect(
                dbname='US_Datamart',
                user=self._bodega_user,
                password=self._bodega_password,
                host=self._bodega_host,
                port=5432,
                sslmode='require'
                )
        except psycopg2.OperationalError as err:
            self._conn = None
            self._logger.error(err)

        return self._conn



"""
    Returns gdrive connection. Note, if no token.pickle or credentials.json
    files are found, OAuth2 flow is promoted to be completed via UI.
"""
import csv
import io
import os
import re
import os.path
import json
import pandas as pd

from hellofresh_data import logging_setup
from hellofresh_data.parameter_store import get_parameter_store_value

from googleapiclient.http import MediaIoBaseDownload
from googleapiclient.errors import HttpError
from googleapiclient.discovery import build
from google.oauth2.service_account import Credentials

PATH_TO_JSON = '/US/OA/Prod/GoogleAPI/service_account_json'

class GoogleDrive():
    """
        Google Drive Helper Class
    """
    def __init__(self):

        self.__location__ = os.getcwd()

        self._logger = logging_setup.init_logging("GoogleDrive")

        self.number_of_rows_pulled = 0
        self.file_name = None

    def get_google_api_service_account(self):
        """
            Returns Google API service account json as string.
        """
        service_account = get_parameter_store_value(PATH_TO_JSON)

        return service_account

    def get_gdrive_connection(self):
        """
            Uses service account json file, pulled from parameter store, to
            authenticate login. No browser login necessary.
        """
        scopes = ['https://www.googleapis.com/auth/drive']

        creds = None

        service_account_str = self.get_google_api_service_account()

        creds = json.loads(service_account_str)
        with open(os.path.join(self.__location__, 'service_account.json'), 'w') as fp:
            json.dump(creds, fp)

        service_account_file = \
        os.path.join(self.__location__, 'service_account.json')

        creds = Credentials.from_service_account_file(service_account_file,
                                                      scopes=scopes)

        return build('drive', 'v3', credentials=creds)

    def get_gsheet_connection(self):
        """
            Uses service account json file, pulled from parameter store, to
            authenticate login. No browser login necessary.
        """
        scopes = ['https://www.googleapis.com/auth/spreadsheets.readonly']

        creds = None

        service_account_str = self.get_google_api_service_account()

        creds = json.loads(service_account_str)
        with open(os.path.join(self.__location__, 'service_account.json'), 'w') as fp:
            json.dump(creds, fp)

        service_account_file = \
        os.path.join(self.__location__, 'service_account.json')

        creds = Credentials.from_service_account_file(service_account_file,
                                                      scopes=scopes)

        return build('sheets', 'v4', credentials=creds)

    def get_gsheet_file_by_id(self, sheet_key, sheet_range):
        """
            Get google sheet data by spreadsheet id
        """
        service = self.get_gsheet_connection()

        sheet = service.spreadsheets()

        try:
            result = sheet.values().get(spreadsheetId=sheet_key,
                                        range=sheet_range).execute()
        except HttpError as err:
            self._logger.warning('No files or sheet found with id: %s',
                                 sheet_key)

        values = result.get('values', [])

        return values

    def camel_to_snake(self, column_name):
        return re.sub('(.)([A-Z][a-z]+)', r'\1_\2', column_name)

    def get_gsheet_file_as_df(self, sheet_key, sheet_range, sheet_header_row,
                              clean_col_name_flag=False):
        """
        Returns sheet id for a given range as a pandas df

        Parameters
        ----------
        sheet_key : str
            id of the google sheet
        sheet_range : str
            name of tab and sheet_range to pull in the following format:
            name_of_the_tab!sheet_rangeFrom:sheet_rangeTo
                i.e. forecast_recipes!A1:G
        sheet_header_row : int
            Integer representation of the header row.
            If very first column of the google sheet is the header
            sheet_header_row=0
        clean_col_name_flag : bool
            If set to True will turn columns into snake case and replace any
            non alphanumetic characters with an underscore:
                "This Is a $Col" -> this_is_a_col

        Returns
        -------
        pandas df
            Pandas representation of google sheet
        """

        data = self.get_gsheet_file_by_id(sheet_key, sheet_range)

        try:
            data_df = pd.DataFrame(data)
        except Exception as err:
            self._logger.error(err)

        new_header = data_df.iloc[sheet_header_row]
        data_df = data_df[sheet_header_row+1:]
        data_df.columns = new_header

        if clean_col_name_flag:
            data_df.columns = [self.camel_to_snake(x) for x in data_df.columns]
            data_df.columns = ([re.sub('[^0-9a-zA-Z]+', '_',
                                       x.lower()) for x in data_df.columns])
            data_df.columns = ([x.replace('__', '_') for x in data_df.columns])
            data_df.columns = ([x.strip('_') for x in data_df.columns])

        return data_df

    def get_gdrive_csv_by_name(self, file_name):
        """
              Search drive by file name and check if file is present.
        """
        self.file_name = file_name
        gdrive_service = self.get_gdrive_connection()

        response = \
        gdrive_service.files().list(q="name='{}'".format(self.file_name),
                                    spaces='drive',
                                    fields='nextPageToken, files(id, name)'
                                    ).execute()

        response_obj = response.get('files', [])

        if not response_obj:
            self._logger.warning('No files found with name: %s', self.file_name)
        else:
            self._logger.info('Found file: %s', response_obj[0])

        data_io = self.get_gdrive_csv_by_id(response_obj[0].get('id'))

        return data_io

    def get_gdrive_csv_by_id(self, sheet_key):
        """
            Search drive by file ID and check if file is present.
            If present, download.
        """
        gdrive_service = self.get_gdrive_connection()

        try:
            response = gdrive_service.files().get_media(fileId=sheet_key)
            fh_io = io.BytesIO()
            downloader = MediaIoBaseDownload(fh_io, response, chunksize=1024*1024)

            self._logger.info('Downloading "%s" from drive...', self.file_name)
            done = False
            while done is False:
                status, done = downloader.next_chunk(num_retries=2)

        except HttpError as err:
            self._logger.error(err)

        self._logger.info('Downloaded "%s" successfully!', self.file_name)

        return fh_io.getvalue()

    def convert_drive_io_data_to_df(self, data):
        """
            Get the data streamed from google drive and convert
            to DataFrame.
        """

        self._logger.info('Convert stream drive data to pandas DataFrame')

        decoded_data = data.decode('utf-8')
        file_io = io.StringIO(decoded_data)
        reader = csv.reader(file_io, delimiter=',')

        data_df = pd.DataFrame(reader)
        data_df = data_df.infer_objects()

        self.number_of_rows_pulled = len(data_df)

        self._logger.info('Pulled %s rows from drive file', self.number_of_rows_pulled)

        return data_df



"""
    parameter_store.py

    Part of the hellofresh_data module meant to
    set up access to AWS' parameter store and
    return values based on the passed in key.
"""

import boto3
from botocore.errorfactory import ClientError

DEFAULT_REGION_NAME = 'us-east-2'

def get_parameter_store_value(for_key):
    """
        Return value for a passed in key
    """
    try:
        client = boto3.client('ssm', region_name=DEFAULT_REGION_NAME)

        response = client.get_parameter(
            Name=for_key,
            WithDecryption=True
        )
        return response.get('Parameter', {}).get('Value')
    except ClientError as err:
        print(err)



"""
    AWS S3 class
"""

import boto3
import io
import os

from botocore.exceptions import ClientError
from hellofresh_data import logging_setup

class S3():
    """
        S3 Helper Class
    """
    def __init__(self):

        self.__location__ = os.getcwd()
        self._logger = logging_setup.init_logging("S3")

    def df_to_s3(self, data_df, s3_bucket_name, file_path):
        """
            Loads data into S3 bucket.

            data_df - pandas DF
            s3_bucket_name - AWS S3 bucket
            file_path - path to file, i.e Folder/SubFolder/file_name.csv

            ex: s3_obj.df_to_s3(df,
                                'hf-us-oa-etl-odl',
                                'IMT/po_master/test_file.csv')
        """

        self._logger.info('Load pandas DataFrame to S3 path, %s/%s',
                          s3_bucket_name,
                          file_path)

        csv_buffer = io.StringIO()
        data_df.to_csv(csv_buffer)

        s3_resource = boto3.resource('s3', region_name="us-east-2")

        try:
            s3_resource.Object(s3_bucket_name,
                               file_path
                               ).put(Body=csv_buffer.getvalue())

            self._logger.info('Loaded "%s" to S3 path, %s',
                              file_path,
                              s3_bucket_name)
        except ClientError as err:
            self._logger.error(err)



"""
    DataMaster
    - Handles reading and writing data from/to Bodega
        DataReader:
            obj_r = DataReader('select * from staging.sasha_hf_zip_maps limit 10;')
            df = obj_r.get_result_of_sql_as_df()
            print(df)
        DataWriter:
            obj = DataWriter(data_df=df,
                             fq_table_name='staging.sasha_delete_this',
                             primary_key_list=['dest_zip_code', 'hellofresh_week'])
            obj.run_load_to_db()

    DataReader only runs "select" statements from tables us_oa_etl user has
    access to.
    DataWriter is meant to write data to tables us_oa_etl user has
    access to. The flow is as follows:
        1. Add/replace timestamp columns
        2. Create intermediary table with the same name as
           the fq_table_name in intermediate_tables schema.
        3. Load data to intermediary table in text only format with no constraints.
        4. Check if production table has primary key. Add passed one if passed
           primary_key_list arguments.
        5. Create prod table, with proper data types/primary keys
           and load from intermediary.
"""
import sys
import io
import os
import yaml
import pandas as pd
from hellofresh_data import logging_setup
from datetime import datetime
from hellofresh_data import bodega_connect
from hellofresh_data.parameter_store import get_parameter_store_value
from psycopg2.errors import InvalidTextRepresentation, BadCopyFileFormat
from psycopg2.extras import RealDictCursor


class DataReader():

    def __init__(self, sql=None):

        self.sql = sql
        self._logger = logging_setup.init_logging('DataReader')

        self.instantiate_config()
        self.instantiate_bodega_creds()
        self.get_connection()
        self.validate_sql()

    def __del__(self):
        try:
            self.bodega_conn.close()
        except AttributeError:
            self.bodega_conn = None

    def instantiate_config(self):
        """
            Instantiate config to expose variables
        """
        self.__location__ = \
            os.path.realpath(
                os.path.join(os.getcwd(), os.path.dirname(__file__)))
        self.config = \
            yaml.safe_load(open(os.path.join(self.__location__, 'config.yml')))

    def instantiate_bodega_creds(self):
        """
            Retrieve bodega credentials for DDL
        """
        bodega_user_path = self.config['data_master']['bodega']['user']
        bodega_password_path = self.config['data_master']['bodega']['password']

        self.db_user = get_parameter_store_value(bodega_user_path)
        self.db_password = get_parameter_store_value(bodega_password_path)

    def get_connection(self):
        """
            Connecting to Bodega
        """

        try:
            conn_obj = bodega_connect.BodegaConnect(self.db_user,
                                                    self.db_password)
            self.bodega_conn = conn_obj.bodega_connect()
        except Exception as err:
            self._logger.error('Failed to connect to DB')
            self._logger.error(err)

    def validate_sql(self):
        """
            Run basic checks on SQL.
            Assert:
                1. Only run `allowed_command` from config.
                2. Only one statement per query, meaning no
                    query delimiters present.

            This is more of a foolproof check vs. checking for an
            actual malicious behaviour. This "ensures" we only use this
            class for reading data.
        """
        sql_parts = self.sql.strip().split(' ')
        delimiter_count = self.sql.strip().count(';')

        allowed_command_ls = \
            self.config['data_master']['sql_validate_reader']['allowed_command']

        if sql_parts[0].lower() not in allowed_command_ls:
            self._logger.error(f'Can only run: {allowed_command_ls}')
            sys.exit(1)

        try:
            delimiter_count = int(delimiter_count)
        except Exception:
            delimiter_count = 0
            pass

        if int(delimiter_count) > 1:
            self._logger.error('Can only execute 1 sql command')
            sys.exit(1)

    def execute_sql(self, sql):
        """
            Execute passed in SQL statement. Rollback on error. Return data if
            available.
        """
        try:
            cur = self.bodega_conn.cursor(cursor_factory=RealDictCursor)
            cur.execute("{sql}".format(sql=sql))
            try:
                rowcount = cur.rowcount
            except Exception:
                pass
            self.bodega_conn.commit()
        except (InvalidTextRepresentation, BadCopyFileFormat):
            self.bodega_conn.rollback()
        except Exception as err:
            self.bodega_conn.rollback()
            self._logger.warning(err)

        try:
            data = cur.fetchall()
        except Exception:
            data = None
            pass

        return data, cur, rowcount

    def get_result_of_sql_as_list(self):
        """
            Return SQL result as a list
        """
        data, cur, rowcount = self.execute_sql(self.sql)
        try:
            col_names = [desc[0] for desc in cur.description]
        except Exception as err:
            self._logger.error(err)
            sys.exit(1)

        return data, col_names

    def get_result_of_sql_as_df(self):
        """
            Return SQL as a Data Frame
        """

        data, col_names = self.get_result_of_sql_as_list()

        return pd.DataFrame(data, columns=col_names)


class DataWriter(DataReader):

    def __init__(self,
                 data_df,
                 fq_table_name,
                 primary_key_list=None,
                 logging_flag=False):

        self.data_df = data_df
        self.fq_table_name = fq_table_name
        self.arg_primary_key_ls = primary_key_list
        self.logging_flag = logging_flag

        self._logger = logging_setup.init_logging('DataWriter')

        super().instantiate_config()
        super().instantiate_bodega_creds()
        super().get_connection()
        self.instantiate_bodega_variables()

    def __del__(self):
        try:
            self.bodega_conn.close()
        except AttributeError:
            self.bodega_conn = None

    def instantiate_bodega_variables(self):
        """
            Initialize DB varibales from config file
        """

        self.prod_table_name = self.fq_table_name.split('.')[1]
        self.prod_schema_name = self.fq_table_name.split('.')[0]

        self.intermediate_schema = \
            self.config['data_master']['bodega']['intermediate_schema']
        self.stage_fq_table_name = '{}.{}'.format(self.intermediate_schema,
                                                  self.prod_table_name)

        try:
            if self.prod_schema_name in self.intermediate_schema:
                self._logger.error('Cannot use %s schema for prod',
                                   self.prod_schema_name)
                sys.exit(1)
        except IndexError:
            self._logger.error('Make sure you pass in schema.table format')
            self._logger.error('You passed: %s', self.fq_table_name)
            sys.exit(1)

    def set_primary_key_variables(self):
        """
            Runs a check on whether `self.fq_table_name` table has a
            primary key, if it does self.current_primary_key_str = True.

            If it does not and self.arg_primary_key_ls were passed in try to
            set primary key to self.arg_primary_key_ls in create_prod_ddl

            If no current primary key and none were passed in log warning and
            proceed.
        """
        self.current_primary_key_str = self.current_primary_key_check()

        if self.current_primary_key_str:
            self.primary_key_exists = True
        elif self.arg_primary_key_ls:
            self.primary_key_exists = False
            self.arg_primary_key_str = ', '.join(self.arg_primary_key_ls)
            self.current_primary_key_str = self.arg_primary_key_str

        if (self.current_primary_key_str is None
           and self.arg_primary_key_ls is None):
            self.primary_key_exists = False
            self._logger.warning(f'Table {self.fq_table_name} does not have '
                                 f'a primary key nor will one be set.')

    def check_if_primary_keys_equal(self):
        """
            If primary key exists and arg_primary_key_ls were passed in
            see if they are equal. Alert in case they are not and proceed with
            current primary key.
        """
        if self.current_primary_key_str and self.arg_primary_key_ls:
            current_primary_key_ls = \
                [x.strip() for x in self.current_primary_key_str.split(',')]

            if set(current_primary_key_ls) != set(self.arg_primary_key_ls):
                self._logger.warning(f'Current primary key list: '
                                     f'{current_primary_key_ls} '
                                     f'does not match what was passed in: '
                                     f'{self.arg_primary_key_ls} '
                                     f'If you want to change primary keys use '
                                     f'you should do it manually i.e DBeaver. '
                                     f'Falling back to Current primary key '
                                     f'list')

    def get_ddl_column_ls(self):
        """
            Create DDL columns
            prod:
                column        text,
                column_number int,
                time_modified timestamp
            stage:
                column         text,
                column_number  text,
                time_modified  text
             stage_to_prod_cast:
                 column::text,
                 column_number::int,
                 time_modified::timestamp
        """
        prod_ddl = []
        stage_ddl = []
        stage_to_prod_cast = []

        df_types = self.data_df.infer_objects().dtypes
        df_column_names = list(self.data_df.columns)

        for i in range(len(df_types)):
            if str(df_types[i]) == 'object':
                data_type = 'text'
            if str(df_types[i]) == 'int64':
                data_type = 'int'
            if str(df_types[i]) == 'float64':
                data_type = 'numeric'
            if str(df_types[i]) == 'bool':
                data_type = 'boolean'
            if str(df_types[i]) == 'datetime64' \
                    or str(df_types[i]) == 'datetime64[ns]':
                data_type = 'timestamp'
            if any(word in df_column_names[i].lower() for
                   word in ['_date', 'date_', ' date']):
                data_type = 'date'

            if ' ' in df_column_names[i]:
                column_name = '"{}"'.format(df_column_names[i])
            else:
                column_name = df_column_names[i]

            prod_ddl.append('{} {}'.format(column_name, data_type))
            stage_ddl.append('{} text'.format(column_name))
            stage_to_prod_cast.append('{}::{}'.format(column_name,
                                                      data_type))

        self.prod_ddl_str = ', '.join(prod_ddl)
        self.stage_ddl_str = ', '.join(stage_ddl)
        self.stage_to_prod_cast_str = ', '.join(stage_to_prod_cast)

    def get_ddl_for_staging_db(self):
        """
            Create a DDL statement to create a table
            based on DF column names.
        """
        if self.logging_flag:
            self._logger.info('Create DDL for staging DB')

        self.staging_ddl_sql = """
                           create table if not exists {fq_table_name}
                          ({stage_ddl_str});
                           """.format(stage_ddl_str=self.stage_ddl_str,
                                      fq_table_name=self.stage_fq_table_name)

    def add_timestamp_columns(self):
        """
            Add created/updated at timestamps in UTC. Drop old ones if exits
        """
        try:
            self.data_df.drop('etl_created_at', 1, inplace=True)
        except Exception:
            pass
        try:
            self.data_df.drop('etl_updated_at', 1, inplace=True)
        except Exception:
            pass

        if self.logging_flag:
            self._logger.info('Adding DF timestamps')

        self.data_df['etl_created_at'] = \
            datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
        self.data_df['etl_updated_at'] = \
            datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')

        self.data_df = self.data_df.astype({'etl_created_at': 'datetime64',
                                            'etl_updated_at': 'datetime64'})

    def create_intermediary_db_table(self):
        """
            Create table from a predefined DDL statement
        """
        if self.logging_flag:
            self._logger.info('Creating table %s', self.stage_fq_table_name)
        super().execute_sql(self.staging_ddl_sql)

    def load_df_to_intermediary(self):
        """
            Loads data into the DB table via COPY command.
            Data is split into chunks and is stored in memory while
            being loaded. Stage table is truncated prior to each exucution.
        """
        if self.logging_flag:
            self._logger.info('Inserting data into %s',
                              self.stage_fq_table_name)

        sio = io.StringIO()
        sio.write(self.data_df.to_csv(index=False,
                                      header=False,
                                      na_rep="[NULL]",
                                      chunksize=10000,
                                      mode='a'))
        sio.seek(0)

        if self.logging_flag:
            self._logger.info('Inserting %s into table %s',
                              len(self.data_df),
                              self.stage_fq_table_name)

        super().execute_sql("""truncate table {}
                            """.format(self.stage_fq_table_name))

        try:
            cur = self.bodega_conn.cursor()
            cur.copy_expert("""
                            COPY {}
                            FROM STDIN
                            WITH (FORMAT CSV, NULL "[NULL]")
                            """.format(self.stage_fq_table_name), sio)
            self.bodega_conn.commit()
        except (InvalidTextRepresentation, BadCopyFileFormat) as err:
            self.bodega_conn.rollback()
            self._logger.error(err)
            sys.exit(1)

        if self.logging_flag:
            self._logger.info('Inserted into bodega table: %s',
                              self.stage_fq_table_name)

    def current_primary_key_check(self):
        """
            Run query from information_schema to check
            if a primary key exists. Return as comma delimited string if one
            exits.
        """

        primary_key = super().execute_sql("""
        select array_to_string(array_agg(kcu.column_name::text), ', ')
               as primary_list
          from information_schema.table_constraints tco
          join information_schema.key_column_usage kcu
            on kcu.constraint_name = tco.constraint_name
           and kcu.constraint_schema = tco.constraint_schema
           and kcu.constraint_name = tco.constraint_name
         where tco.constraint_type = 'PRIMARY KEY'
           and kcu.table_name = '{table_name}'
           and kcu.table_schema = '{schema_name}'
                   """.format(table_name=self.prod_table_name,
                              schema_name=self.prod_schema_name))
        try:
            current_primary_key_str = primary_key[0][0]['primary_list']
        except IndexError:
            current_primary_key_str = None

        return current_primary_key_str

    def check_if_table_exist(self):
        """
            Return prod_exist_flag, used for primary key creation.
        """

        prod_exist_check = super().execute_sql("""
           SELECT table_name as table_name
             FROM information_schema.tables
            WHERE table_schema = '{schema_name}'
              AND table_name   = '{table_name}'
        """.format(table_name=self.prod_table_name,
                   schema_name=self.prod_schema_name))
        try:
            prod_exist_check[0][0]['table_name']
            prod_exist_flag = True
        except IndexError:
            prod_exist_flag = False

        return prod_exist_flag

    def create_prod_ddl(self):
        """
            DDL for prod table

            For primary key, only add one if:
                1. Currenly there is not primary key on the table
                2. Prod table does not currenly exist
                3. Primary key argument was passed in

            Above conditions have to ALL be met to set a primary key. Logic for
            #2 is that if a table was created without a primary key, one should
            not be set using DataWriter.
        """
        self.prod_ddl_ls = []

        prod_create_ddl_sql = """
             create table
             if not exists {fq_table_name}
                           ({prod_ddl_str});
           """.format(prod_ddl_str=self.prod_ddl_str,
                      fq_table_name=self.fq_table_name)

        if (self.primary_key_exists is False and
           self.check_if_table_exist() is False and
           self.arg_primary_key_ls):
            prod_create_primary_key_sql = """
                                    alter table {fq_table_name}
                                    add primary key ({constraint});
                                """.format(fq_table_name=self.fq_table_name,
                                           constraint=self.arg_primary_key_str)
        elif (self.check_if_table_exist() is True and self.arg_primary_key_ls):
            self._logger.warning(f'You passed in primary_key_list, however, '
                                 f'since {self.fq_table_name} already exists, '
                                 f'these will not be added as primary keys. '
                                 f'If needed, add manually via DBeaver.')
            prod_create_primary_key_sql = None
        else:
            prod_create_primary_key_sql = None

        prod_grant_ddl_sql = """
             grant all privileges on table {fq_table_name} to root;
             grant select on table {fq_table_name} to analyst;
             grant select on table {fq_table_name} to sashap;
                """.format(fq_table_name=self.fq_table_name)

        self.prod_ddl_ls.extend([prod_create_ddl_sql,
                                 prod_create_primary_key_sql,
                                 prod_grant_ddl_sql])

    def create_filter_for_prod_insert(self):
        """
            Generates the filter clause for the intermediary to prod insert
            based on primary key(s).
            i.e
                where stage.a = prod.a
                  and stage.b = prod.b --If composite primary key

        """

        pm_list = self.current_primary_key_str.split(', ')
        filter_str = ''
        if len(pm_list) > 1:
            filter_str += \
                f'where stage.{pm_list[0]}::text = prod.{pm_list[0]}::text'
            for key in pm_list[1:]:
                filter_str += (' and ')
                filter_str += (f'stage.{key}::text = prod.{key}::text ')
        else:
            filter_str += \
                f'where stage.{pm_list[0]}::text = prod.{pm_list[0]}::text'

        return filter_str.strip()

    def create_prod_insert_ddl(self):
        """
            Generate insert into prod table from intermediary.
        """

        if self.primary_key_exists:
            if_primary_key_exist_sql = """
                        where not exists
                              (select 1
                                 from {prod_fql} prod
                                {filter_clause}
                               )
                           on conflict ({constraint})
                           do nothing;
            """.format(prod_fql=self.fq_table_name,
                       filter_clause=self.create_filter_for_prod_insert(),
                       constraint=self.current_primary_key_str)
        else:
            if_primary_key_exist_sql = ''

        self.prod_insert_ddl_sql = """
                    insert into {prod_fql}
                    select {stage_to_prod_cast_str}
                      from {stage_fql} stage
                      {if_primary_key_exist_sql}
              """.format(prod_fql=self.fq_table_name,
                         stage_to_prod_cast_str=self.stage_to_prod_cast_str,
                         stage_fql=self.stage_fq_table_name,
                         if_primary_key_exist_sql=if_primary_key_exist_sql)

    def execute_ddl_commads(self):
        """
            Executes DDL statements generated in create_prod_ddl()
        """

        for sql in self.prod_ddl_ls:
            if sql:
                if self.logging_flag:
                    self._logger.debug(sql)
                    super().execute_sql(sql)
                else:
                    super().execute_sql(sql)

    def execute_load_to_prod_command(self):
        """
            Execute DDL statement generated in create_prod_insert_ddl()
        """

        if self.logging_flag:
            self._logger.debug(self.prod_insert_ddl_sql)
            data, cur, rowcount = super().execute_sql(self.prod_insert_ddl_sql)
        else:
            data, cur, rowcount = super().execute_sql(self.prod_insert_ddl_sql)

        self._logger.info(f'Inserted {rowcount} record(s) into Prod DB: '
                          f'{self.fq_table_name}')

    def run_load_to_db(self):
        """
            Run all functions
        """
        self.set_primary_key_variables()
        if self.primary_key_exists:
            self.check_if_primary_keys_equal()
        self.add_timestamp_columns()
        self.get_ddl_column_ls()
        self.get_ddl_for_staging_db()
        self.create_intermediary_db_table()
        self.load_df_to_intermediary()
        self.create_prod_ddl()
        self.execute_ddl_commads()
        self.create_prod_insert_ddl()
        self.execute_load_to_prod_command()



"""
    Reader class for DWH

    Default connection is using sca service account's credentials
"""
import sys
import os
import yaml
import pandas as pd
from impala.dbapi import connect
from hellofresh_data import logging_setup
from hellofresh_data.parameter_store import get_parameter_store_value


class ImpalaReader():

    def __init__(self, user=None, password=None):
        self.__user_var = user
        self.__password_var = password
        self._logger = logging_setup.init_logging('ImpalaReader')
        self.instantiate_config()
        self.instantiate_impala_creds()
        self.__get_connection()

    def __del__(self):
        try:
            self.__impala_conn.close()
        except Exception:
            self.__impala_conn = None

    def instantiate_config(self):
        """
            Instantiate config to expose variables
        """
        self.__location__ = \
            os.path.realpath(
                os.path.join(os.getcwd(), os.path.dirname(__file__)))
        self.config = \
            yaml.safe_load(open(os.path.join(self.__location__, 'config.yml')))

    def instantiate_impala_creds(self):
        """
            Set DWH variables

            For self.__user and self.__password either use passed in variables
            or assign those of a sca service account
        """
        __user = self.config['impala_reader']['dwh']['user']
        __password = self.config['impala_reader']['dwh']['password']

        self.__user = self.__user_var or get_parameter_store_value(__user)
        self.__password = \
            self.__password_var or get_parameter_store_value(__password)

        self._logger.info(f'Using creds for user: {self.__user}')

        self.host = self.config['impala_reader']['dwh']['host']
        self.port = self.config['impala_reader']['dwh']['port']
        self.auth_mechanism = \
            self.config['impala_reader']['dwh']['auth_mechanism']
        self.use_ssl = self.config['impala_reader']['dwh']['use_ssl']

    def __get_connection(self):
        """
            Try to establish connection to DWH using set variables
        """

        try:
            self.__impala_conn = connect(host=self.host,
                                         port=self.port,
                                         user=self.__user,
                                         password=self.__password,
                                         auth_mechanism=self.auth_mechanism,
                                         use_ssl=self.use_ssl)
        except Exception as err:
            self._logger.error('Failed to connect to DWH')
            self._logger.error(err)

    def validate_sql(self, sql):
        """
            Run basic checks on SQL.
            1. First command of SQL must be in the "allowed_command_ls" list
            2. Commands in forbidden_command_ls cannot exist in SQL
            3. Only one statement can be executed at a time

            This is more of a foolproof check vs. checking for an
            actual malicious behaviour. This "ensures" we only use this
            class for reading/refreshing data.
        """
        sql_parts = sql.strip().split(' ')
        delimiter_count = sql.strip().count(';')

        allowed_command_ls = \
            self.config['impala_reader']['sql_validate']['allowed_command']

        if sql_parts[0].lower() not in allowed_command_ls:
            self._logger.error(f'Can only run: {allowed_command_ls}')
            sys.exit(1)

        forbidden_command_ls = \
            self.config['impala_reader']['sql_validate']['forbidden_command']

        if any(ext in sql_parts for ext in forbidden_command_ls):
            self._logger.error(f'SQL cannot include: {forbidden_command_ls}')
            sys.exit(1)

        try:
            delimiter_count = int(delimiter_count)
        except Exception:
            delimiter_count = 0
            pass

        if int(delimiter_count) > 1:
            self._logger.error('Can only execute 1 sql command at a time')
            sys.exit(1)

    def execute_sql(self, sql):
        """
            Validate passed in query and then try to execute without return
            value. This function is meant to run "refresh"
            and "invalidate metadata" statements.
        """
        self.validate_sql(sql)

        cursor = self.__impala_conn.cursor()
        try:
            cursor.execute(sql)
            self._logger.info(f'SUCCESS: {sql}')
        except Exception as err:
            self._logger.error(err)
            sys.exit(1)

    def get_result_of_sql_as_df(self, sql):
        """
            Validate passed in query and then try to execute and return as a
            pandas DataFrame. This function is meant to run "select"
            statements.
        """
        self.validate_sql(sql)

        try:
            df = pd.read_sql(sql, self.__impala_conn)
        except Exception as err:
            df = None
            self._logger.error(err)

        return df



import setuptools

setuptools.setup(
    name="hellofresh_oa_data",
    version="0.3.11",
    author="US OA Team",
    author_email="sca-automation@hellofresh.com",
    long_description="Package with helper function to support functions of OA team in regarding to data",
    url="https://github.com/hellofresh/us-ops-tech-scripts/hellofresh-oa-data",
    packages=setuptools.find_packages(),
    zip_safe=False,
    package_data={'': ['config.yml']},
    include_package_data=True
)



"""
    logging_setup.py

    Part of the hellofresh_data module meant to
    return a logger object in order to log various
    events.
"""
import logging


def init_logging(logger_name, loglevel=logging.DEBUG):
    """
        Creates standard logging for the logger_name passed in
    """

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    channel = logging.StreamHandler()
    channel.setLevel(loglevel)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    channel.setFormatter(formatter)

    logger.addHandler(channel)
    logger.debug("Initialized logging for %s", logger_name)

    return logger
